from .whatsapp.whatsapp import Whatsapp
from .whatsapp.markup import Inline_button, Inline_keyboard, Inline_list, List_item
from .whatsapp.user_context import _context
from .whatsapp.dispatcher import Update, Message_handler, Interactive_query_handler
print("importedddddd")



class _Context:
    """Object to store all users data in a conversation"""

    def __init__(self) -> None:
        self.users_data = {}

    def _add_user(self, phone_num):
        self.users_data[str(phone_num)] = {}

    def _user_exists(self, phone_num):
        if len(self.users_data):
            if phone_num in self.users_data.keys():
                return True


_context = _Context()


class User_context():
    """Object that manages a specific user's data in a conversation.
    The user's phone number id is used as the id"""

    def __init__(self, phone_num: str) -> None:
        if not _context._user_exists(phone_num):
            _context._add_user(phone_num)
        self.user_data = _context.users_data[str(phone_num)]

    def user_data(self):
        pass



from typing import Union


class Reply_markup:
    def __init__(self, markup) -> None:
        self.markup = markup
    # if item is for keyboard, initialize with button settings
    # if item is for list actions, initialize with list settings


class Inline_button:
    def __init__(self, text: str, button_id: str = None):

        self.button = {
            "type": "reply",
                    "reply": {
                        "id": button_id if button_id else text,
                        "title": text
                    }
        }
    __slots__ = ('button')


class Inline_keyboard(Reply_markup):
    """Accepts only three(3) text (or buttons) in a flat list.
    Minimum of one(1)"""

    def __init__(self, inline_buttons: Union[list[str], list[Inline_button]]):
        self.inline_buttons = self.set_buttons(inline_buttons)
        self.error_check()
        self.keyboard = self.set_keys()
        super().__init__(self.keyboard)

    def set_buttons(self, _buttons: Union[list[str], list[Inline_button]]):
        if not isinstance(_buttons, list):
            raise ValueError("List argument expected")
        res = []
        for i in _buttons:
            if isinstance(i, str):
                res.append(Inline_button(i))
            elif isinstance(i, Inline_button):
                res.append(i)
            else:
                raise ValueError(
                    "str or Inline_button expected as button elements")
        return res

    def error_check(self):
        if len(self.inline_buttons) > 3 or len(self.inline_buttons) < 1:
            raise ValueError(
                "Inline_keyboard can only accept minimum of 1 Inline_button item and maximum of 3")
        button_id_check = []
        button_text_check = []
        for i, button in enumerate(self.inline_buttons):
            if not isinstance(button, Inline_button):
                raise ValueError(
                    f"Item at position {i} of list argument expected to be string or an instance of Inline_button")
            butt = button.button['reply']['id']
            buttt = button.button['reply']['title']
            if butt in button_id_check or buttt in button_text_check:
                raise ValueError("Use unique id and text for the buttons")
            button_id_check.append(butt)
            button_text_check.append(buttt)

    def set_keys(self):
        action = {"buttons": [i.button for i in self.inline_buttons]}
        return action


class List_item():
    def __init__(self, title: str, _id: str = None, description: str = None) -> None:
        self.title = title
        self._id = _id if _id else self.title
        self.item = {
            "id": self._id,
            "title": self.title
        }
        if description:
            self.item["description"] = description
    __slots__ = ('title', 'item', '_id')


class List_section():
    def __init__(self, title: str, items_list: Union[list[str], list[List_item]]) -> None:
        self.title = title
        self.items_list = self.set_list(items_list)
        self.error_check()
        self.section = self.set_section()

    def set_list(self, _items_list: Union[list[str], list[List_item]]):
        if not isinstance(_items_list, list):
            raise ValueError("List argument expected")
        res = []
        for i in _items_list:
            if isinstance(i, str):
                res.append(List_item(i))
            elif isinstance(i, List_item):
                res.append(i)
            else:
                raise ValueError(
                    "str or List_item object expected as items_list elements")
        return res

    def error_check(self):
        for i, item in enumerate(self.items_list):
            if not isinstance(item, List_item):
                raise ValueError(
                    f"Item at position {i} of list argument expected to be an instance of Inline_button")

    def set_section(self):
        sections = {"title": self.title,
                    "rows": [i.item for i in self.items_list]}
        return sections


class Inline_list(Reply_markup):
    """Accepts up to ten(10) list items. Minimum of one(1)
    Accepts one level of nesting, e.g [[],[],[]].
    Args:
        button_text: (str),required - Specifies the button that displays the list items when clicked
        list_items:(list) required - Specifies the list items to be listed. Maximum of ten items. 
            These Items are defined with the List_item class.
            To use sections, pass a list of List_section instances instead"""

    def __init__(self, button_text: str, list_items: Union[list[List_item], list[List_section]]):
        self.button_text = button_text
        self.list_items = list_items
        self.error_check()
        self.inline_list = self.set_list()
        super().__init__(self.inline_list)

    def error_check(self):
        if not isinstance(self.list_items, list):
            raise ValueError(
                "The argument for list_items should be of type 'list'")

        for i, item in enumerate(self.list_items):
            if not (isinstance(item, List_item) or isinstance(item, List_section)):
                # Check that non-nested list is List_item
                raise ValueError(
                    f"Item at position {i} of list argument expected to be an instance of List_item or list of List_section")

    def set_list(self):
        action = {
            "button": self.button_text,
            "sections": [i.section for i in self.list_items] if isinstance(self.list_items, List_section) else [{"rows": [i.item for i in self.list_items]}]}
        return action



import inspect
from typing import Callable
from .markup import Reply_markup
from .error_handlers import keys_exists
import re
from .user_context import User_context
#from queue import Queue
from threading import Thread


class Update:
    def __init__(self, bot, update) -> None:
        self.bot = bot
        self.value = update
        try:
            self.message = self.value["messages"][0]
        except:
            self.message = {}
        self.user = self.value["contacts"][0]
        self.user_display_name = self.user["profile"]["name"]
        self.user_phone_number = self.user["wa_id"]
        self.message_text = None
        #self.list_text = None
        self.interactive_text = None
        if keys_exists(self.message, "text", "body"):
            self.message_text = self.message["text"]["body"]
        if keys_exists(self.message, "interactive", "list_reply"):
            self.interactive_text = self.message["interactive"]["list_reply"]
            self.message_text = self.message["interactive"]["list_reply"]['id']
        if keys_exists(self.message, "interactive", "button_reply"):
            self.interactive_text = self.message["interactive"]["button_reply"]
            self.message_text = self.message["interactive"]["button_reply"]['id']

    def reply_message(self, text: str, reply_markup: Reply_markup = None, header: str = None, header_type: str = "text", footer: str = None, web_page_preview=True, *args, **kwargs):
        self.bot.send_message(self.user_phone_number, text, reply_markup=reply_markup,
                              header=header, header_type=header_type, footer=footer, web_page_preview=web_page_preview, *args, **kwargs)

    def reply_media(self, media_path, caption: str = None, media_provider_token: str = None, *args, **kwargs):
        self.bot.send_media_message(
            self.user_phone_number, media_path, caption, media_provider_token, *args, **kwargs)


class Update_handler:
    def __init__(self, context: bool = True) -> None:
        self.name = None
        self.regex = None
        self.func = None
        self.action = None
        self.context = context
        self.list = None
        self.button = None
        self.persistent = False

    def filter_check(self, msg) -> bool:
        if self.regex:
            if re.match(self.regex, msg):
                return True
            else:
                return False
        if self.func:
            if self.func(msg):
                return True
            else:
                return False

        return True

    def run(self, *args, **kwargs):
        new_kwargs = {
            i: kwargs[i] for i in kwargs if i in inspect.getfullargspec(self.action).args}
        return self.action(*args, **new_kwargs)


class Message_handler(Update_handler):
    def __init__(self, regex: str = None, func: Callable = None, action: Callable = None, context: bool = True, persistent: bool = False) -> None:
        super().__init__(context)
        self.name = "text"
        self.regex = regex
        self.func = func
        self.action = action
        self.persistent = persistent


class Interactive_query_handler(Update_handler):
    """For button_reply and list_reply"""

    def __init__(self, regex: str = None, func: Callable = None, handle_button: bool = True, handle_list: bool = True, action: Callable = None, context: bool = True, persistent: bool = False) -> None:
        super().__init__(context)
        self.name = "interactive"
        self.regex = regex
        self.func = func
        self.action = action
        self.list = handle_list
        self.button = handle_button
        self.persistent = persistent


class Location_handler(Update_handler):
    def __init__(self) -> None:
        super().__init__()


class Media_handler(Update_handler):
    """Image, audio and video"""

    def __init__(self) -> None:
        super().__init__()


class Dispatcher:
    def __init__(self, bot, mark_as_read: bool = True) -> None:
        self.bot = bot
        self.queue = bot.queue
        self.registered_handlers = []  # list of handler instances
        self.mark_as_read = mark_as_read
        self.next_step_handler = {}
        self.fallback_function = None

    def process_update(self, update) -> None:
        # ------------new---------
        # self._process_queue(update)
        # ------------old3--------
        # Thread(target=self._process_queue(update)).start()

        # -----------old2------
        # if not self.bot.threaded:
        #     self._process_queue(update)
        # else:
        #     Thread(target=self._process_queue(update)).start()

        # ---------------old------
        self.queue.put(update)
        while True:
            _update = self.queue.get()
            if not self.bot.threaded:
                self._process_queue(_update)
            else:
                Thread(target=self._process_queue(_update)).start()
            if self.queue.empty():
                break

    def _process_queue(self, update) -> None:
        if not keys_exists(update, "entry", 0, "changes", 0, "value"):
            return
        value = update["entry"][0]["changes"][0]["value"]
        #self.value = value
        if not keys_exists(value, "metadata", "phone_number_id"):
            return
        if value["metadata"]["phone_number_id"] == self.bot.id:
            if not keys_exists(value, "messages"):
                return
            _message = value["messages"][0]
            if self.mark_as_read:
                self.bot.mark_as_read(_message)
            update = Update(self.bot, value)
            print(update.message)

            # check if a next step handler has been registered
            persistent_handlers = [
                i for i in self.registered_handlers if i.persistent]
            try:
                users_next_step = self.next_step_handler[update.user_phone_number]
                users_next_step_handler = users_next_step['next_step_handler']
                matched_handlers = []
                try:
                    users_next_step_fallback = users_next_step['fallback_function']
                    matched_handlers.append(users_next_step_fallback)
                except KeyError:
                    pass
                matched_handlers.append(users_next_step_handler)
            # get registered handlers if no next step handler
            except KeyError:
                matched_handlers = [i for i in self.registered_handlers]
            # print(len(persistent_handlers), len(matched_handlers))
            matched_handlers = persistent_handlers+matched_handlers
            for handler in matched_handlers:
                # print(len(persistent_handlers), len(matched_handlers))
                # print(handler.name, handler.regex)
                if (isinstance(handler, Update_handler) and handler.name == _message["type"]) or isinstance(handler, Update_handler):

                    # handle text
                    message_txt = ''
                    if _message["type"] == "text":
                        message_txt = _message["text"]["body"]

                    # handle interactive
                    elif _message["type"] == "interactive":
                        if _message["interactive"]["type"] == 'button_reply' and handler.button:
                            message_txt = _message["interactive"]['button_reply']['id']
                        elif _message["interactive"]["type"] == 'list_reply' and handler.list:
                            message_txt = _message["interactive"]['list_reply']['id']
                        # else:
                        #     return
                    res = self._check_and_run_handler(
                        handler, value, message_txt)
                    if res:
                        try:
                            if self.next_step_handler[update.user_phone_number]['next_step_handler'] == handler or self.next_step_handler[update.user_phone_number]['fallback_function'] == handler:
                                del self.next_step_handler[update.user_phone_number]
                            return
                        except KeyError:
                            # print("key error", handler.regex)
                            return
                    else:
                        continue

    def _check_and_run_handler(self, handler, value, message):
        if hasattr(handler, 'filter_check'):
            if not handler.filter_check(message):
                return False
            if handler.context:
                update = Update(self.bot, value)
                handler.run(update, context=User_context(
                    update.user_phone_number))
            else:
                handler.run(update)
            return True
        return False

    def add_handler(self, handler_instance):
        self.registered_handlers.append(handler_instance)
        handler_index = len(self.registered_handlers)-1
        return handler_index

    def set_next_handler(self, update: Update, function: Callable, handler_type: Update_handler = Update_handler, regex: str = None, func: Callable = None, end_conversation_action: Callable = lambda x: x, end_conversation_keyword_regex: str = r"(?i)^(end|stop|cancel)$"):
        """Sets a function for handling of next update. 
        The set_next_handler overrides other handlers till it handles an update itself"""
        if not issubclass(handler_type, Update_handler):
            return "type should be an Update_handler class"
        regex = regex
        func = func
        self.next_step_handler[update.user_phone_number] = {}
        users_next = self.next_step_handler[update.user_phone_number]
        users_next['fallback_function'] = Message_handler(
            regex=end_conversation_keyword_regex, action=end_conversation_action)
        if handler_type == Message_handler:
            users_next['next_step_handler'] = Message_handler(
                regex, func, action=function)
        elif handler_type == Interactive_query_handler:
            users_next['next_step_handler'] = Interactive_query_handler(
                regex, func, action=function)
        else:
            try:
                # _type = update["entry"][0]["changes"][0]["value"]["messages"][0]["type"]
                _type = update.value["messages"][0]["type"]
                print("new handler", _type)
                new_handler = Update_handler()
                # new_handler.name = _type
                new_handler.regex = regex
                new_handler.func = func
                new_handler.action = function
                users_next['next_step_handler'] = new_handler
            except KeyError:
                return

    def add_message_handler(self, regex: str = None, func: Callable = None, context: bool = True, persistent: bool = False):
        def inner(function):
            _handler = Message_handler(
                regex=regex, func=func, action=function, context=context, persistent=persistent)
            self.add_handler(_handler)
            return function
        return inner

    def add_interactive_handler(self, regex: str = None, func: Callable = None, handle_button: bool = True, handle_list: bool = True, action: Callable = None, context: bool = True, persistent: bool = False):
        def inner(function):
            _handler = Interactive_query_handler(
                regex=regex, func=func, action=function, handle_button=handle_button, handle_list=handle_list, context=context, persistent=persistent)
            self.add_handler(_handler)
            return function
        return inner

    def location_handler(self):
        pass

    def media_handler(self):
        pass



from .dispatcher import Dispatcher, Update
from .message import (
    message_interactive, mark_as_read, message_text, message_template, upload_media, message_media, message_location)
from .markup import Reply_markup
from typing import Union
from queue import Queue


class Whatsapp():
    version_number = 14

    def __init__(self, number_id: int, token: str, mark_as_read: bool = True) -> None:
        """This is the main Whatsapp class. Use it to initialize your bot
        Args:
            id: Your phone number id provided by WhatsApp cloud
            token : Your token provided by WhatsApp cloud
            mark_as_read:(bool), Use to set whether incoming messages should be marked as read. Default is True"""
        self.queue = Queue()
        self.threaded = True
        self.id = number_id
        self.token = token
        self.base_url = f"https://graph.facebook.com/v{str(float(self.version_number))}/{str(self.id)}"
        self.msg_url = self.base_url+"/messages"
        self.media_url = self.base_url+'/media'
        self.dispatcher = Dispatcher(self, mark_as_read)
        self.on_message = self.dispatcher.add_message_handler
        self.on_interactive_message = self.dispatcher.add_interactive_handler
        self.set_next_step = self.dispatcher.set_next_handler

    def process_update(self, update):
        return self.dispatcher.process_update(update)

    def mark_as_read(self, update):
        """Mark any message as read"""
        return mark_as_read(update, self.msg_url, self.token)

    def reply_message(self, update: Update, text, reply_markup: Reply_markup = None, header: str = None, header_type: str = "text", footer: str = None, web_page_preview=True):
        return self.send_message(
            update.user_phone_number, text, reply_markup, header, header_type, footer, web_page_preview=web_page_preview)

    def reply_template(self, update: Update, template_name: str):
        return self.send_template_message(
            update.user_phone_number, template_name)

    def reply_media(self, update: Update, image_path: str, caption: str = None, media_provider_token: str = None):
        return self.send_media_message(
            update.user_phone_number, image_path, caption, media_provider_token)

    def send_message(self, phone_num: str, text: str, reply_markup: Reply_markup = None, header: str = None, header_type: str = "text", footer: str = None, web_page_preview=True):
        """Sends text message
        Args:
            phone_num:(int) Recipeint's phone number
            text:(str) The text to be sent
            web_page_preview:(bool),optional. Turn web page preview of links on/off
        """
        if reply_markup:
            return message_interactive(self.msg_url, self.token, phone_num, text, reply_markup, header=header, header_type=header_type, footer=footer, web_page_preview=web_page_preview)
        else:
            return message_text(self.msg_url, self.token, phone_num, text, web_page_preview=web_page_preview)

    def send_template_message(self, phone_num: str, template_name: str):
        """Sends preregistered template message"""
        return message_template(self.msg_url, self.token, phone_num, template_name)

    def upload_media(self,):

        return upload_media(self.media_url)

    def send_media_message(self, phone_num: str, image_path: str, caption: str = None, media_provider_token: str = None):
        """Sends media message which may include audio, document, image, sticker, or video 
        Using media link or by uploading media from file.
        paths starting with http(s):// or www. will be treated as link, others will be treated as local files """
        return message_media(self.msg_url, self.token, phone_num, image_path, caption, media_provider_token)



def keys_exists(element, *keys):
    '''
    Check if *keys (nested) exists in `element` (dict). e.g:
    data = {
    "spam": {
        "egg": {
            "bacon": "Well..",
            "sausages": "Spam egg sausages and spam",
            "spam": "does not have much spam in it"
        }
    }
    }
    keys_exists(data, "spam") -> True
    keys_exists(data, "spam", "bacon") -> False
    keys_exists(data, "spam", "egg") ->True
    Keys_exists(data, "spam", "egg", "bacon") ->True
    '''
    if not isinstance(element, dict):
        raise AttributeError('keys_exists() expects dict as first argument.')
    if len(keys) == 0:
        raise AttributeError(
            'keys_exists() expects at least two arguments, one given.')

    _element = element
    for key in keys:
        try:
            _element = _element[key]
        except KeyError:
            return False
    return True



import re
from typing import Union
import json
import requests
from .markup import Reply_markup, Inline_button, Inline_keyboard, Inline_list, List_item, List_section


def headers(WA_TOKEN):
    return {
        'Content-Type': 'application/json',
        'Authorization': f"Bearer {WA_TOKEN}"}


def mark_as_read(update, url: str, token: str):
    payload = json.dumps({"messaging_product": "whatsapp",
                          "status": "read",
                          "message_id": update["id"]
                          })
    response = requests.post(url, headers=headers(token), data=payload)
    # print(response.text)
    return response


def message_text(url: str, token: str, phone_num: str, text: str, web_page_preview=True):
    payload = json.dumps({
        "messaging_product": "whatsapp",
        "to": str(phone_num),
        "recipient_type": "individual",
        "type": "text",
        "text": {
            "body": text,
            "preview_url": web_page_preview}})
    response = requests.post(url, headers=headers(token), data=payload)
    # print(response.text)
    return response


def message_interactive(url: str, token: str, phone_num: str, text: str, reply_markup: Reply_markup, header: str = None, header_type: str = "text", footer: str = None, web_page_preview=True):
    if not isinstance(reply_markup, Reply_markup):
        raise ValueError("Reply markup must be a Reply_markup object")
    message_frame = {
        "messaging_product": "whatsapp",
        "to": str(phone_num),
        "recipient_type": "individual",
        "type": "interactive",
        "interactive": {
            "type": _get_markup_type(reply_markup),
            "body": {
                "text": text
            },
            "action": reply_markup.markup}}
    if header:
        if header_type == 'text':
            message_frame["interactive"]["header"] = {
                "type": 'text',  # [text,video,image,document]
                "text": header
            }
        elif header_type in ['image', 'video', 'document']:
            if re.match(r"^((http[s]?://)|(www.))", header):
                header_type_object = {'link': header}
            else:
                header_type_object = {'id': header}
            message_frame["interactive"]["header"] = {
                "type": header_type,  # [text,video,image,document]
                header_type: header_type_object
            }

    if footer:
        message_frame["interactive"]["footer"] = {
            "text": footer
        }
    payload = json.dumps(message_frame)
    response = requests.post(url, headers=headers(token), data=payload)
    return response


def message_template(url: str, token: str, phone_num: str, template_name: str):
    payload = json.dumps({
        "messaging_product": "whatsapp",
        "to": str(phone_num),
        "recipient_type": "individual",
        "type": "template",
                "template": {
                    "name": template_name,
                    "language": {
                        "code": "en_US"
                    }}})
    response = requests.post(url, headers=headers(token), data=payload)
    # print(response.text)
    return response


def _get_markup_type(markup):
    if isinstance(markup, Inline_keyboard):
        return "button"
    elif isinstance(markup, Inline_list):
        return "list"


def upload_media(url: str, token: str,):
    payload = json.dumps({
        "messaging_product": "whatsapp",
    })
    response = requests.post(url, headers=headers(token), data=payload)
    # print(response.text)
    return response


def message_media(url: str, token: str, phone_num: str, image_path: str, caption: str = None, media_provider_token: str = None):
    payload = json.dumps({
        "messaging_product": "whatsapp",
        "to": str(phone_num),
        "recipient_type": "individual",
        "type": "image",
        "image": {
            # "id" : "MEDIA-OBJECT-ID"
            "link": image_path,
            "caption": caption
        }})
    print("kkkiiiimmmmggg", image_path, phone_num)
    response = requests.post(url, headers=headers(token), data=payload)
    # print(response.text)
    return response


def message_location(url: str, token: str, phone_num: str, text: str, web_page_preview=True):
    payload = json.dumps({
        "messaging_product": "whatsapp",
        "to": str(phone_num),
        "recipient_type": "individual",
        "type": "text",
        "text": {
            "body": text,
            "preview_url": web_page_preview}})
    response = requests.post(url, headers=headers(token), data=payload)
    # print(response.text)
    return response
